@@ -1,76 +1,123 @@
 """XMPP bots for humans."""
 
+import re
 from argparse import ArgumentParser
 from configparser import ConfigParser
 from datetime import datetime as dt
 from getpass import getpass
 from imghdr import what
+from inspect import cleandoc
 from logging import DEBUG, INFO, basicConfig, getLogger
 from os import environ
 from os.path import exists
 from pathlib import Path
+from random import choice
 from sys import exit, stdout
 
 from humanize import naturaldelta
+from redis import Redis
 from slixmpp import ClientXMPP
 
 
 class SimpleMessage:
     """A simple message interface."""
 
-    def __init__(self, message):
+    def __init__(self, message, bot):
+        """Initialise the object."""
         self.message = message
+        self.bot = bot
 
     @property
-    def body(self):
+    def text(self):
+        """The entire message text."""
         return self.message["body"]
 
+    @property
+    def content(self):
+        """The content of the message received.
+
+        This implementation aims to match and extract the content of the
+        messages directed at bots in group chats. So, for example, when sending
+        messages like so.
+
+        echobot: hi
+        echobot, hi
+        echobot  hi
+
+        The result produced by `message.content` will always be "hi". This
+        makes it easier to work with various commands and avoid messy parsing
+        logic in end-user implementations.
+        """
+        body = self.message["body"]
+
+        try:
+            match = f"^{self.bot.nick}.?(\s)"
+            split = re.split(match, body)
+            filtered = list(filter(None, split))
+            return filtered[-1].strip()
+        except Exception as exception:
+            self.bot.log.error(f"Couldn't parse {body}: {exception}")
+            return None
+
     @property
     def sender(self):
+        """The sender of the message."""
         return self.message["from"]
 
     @property
     def room(self):
+        """The room from which the message originated."""
         return self.message["from"].bare
 
     @property
     def receiver(self):
+        """The receiver of the message."""
         return self.message["to"]
 
     @property
-    def nickname(self):
-        return self.message["mucnick"]
+    def type(self):
+        """The type of the message."""
+        return self.message["type"]
 
     @property
-    def type(self):
-        return self.message["type"]
+    def nick(self):
+        """The nick of the message."""
+        return self.message["mucnick"]
 
 
 class Config:
     """Bot file configuration."""
 
     def __init__(self, name, config):
+        """Initialise the object."""
         self.name = name
         self.config = config
         self.section = config[self.name] if self.name in config else {}
 
     @property
     def account(self):
+        """The account of the bot."""
         return self.section.get("account", None)
 
     @property
     def password(self):
+        """The password of the bot account."""
         return self.section.get("password", None)
 
     @property
     def nick(self):
+        """The nickname of the bot."""
         return self.section.get("nick", None)
 
 
 class Bot(ClientXMPP):
     """XMPP bots for humans."""
 
+    DIRECT_MESSAGE_TYPES = ("chat", "normal")
+    GROUP_MESSAGE_TYPES = ("groupchat", "normal")
+
     def __init__(self):
+        """Initialise the object."""
         self.name = type(self).__name__.lower()
         self.start = dt.now()
 
@@ -82,6 +129,7 @@ class Bot(ClientXMPP):
         self.init_bot()
         self.register_xmpp_event_handlers()
         self.register_xmpp_plugins()
+        self.init_db()
         self.run()
 
     def parse_arguments(self):
@@ -205,26 +253,23 @@ class Bot(ClientXMPP):
         self.add_event_handler("message_error", self.error_message)
 
     def error_message(self, message):
-        _message = SimpleMessage(message)
-        self.log.error(f"Received error message: {_message.body}")
+        message = SimpleMessage(message, self)
+        self.log.error(f"Received error message: {message.text}")
 
     def direct_message(self, message):
-        """Handle message event."""
-        if message["type"] in ("chat", "normal"):
-            _message = SimpleMessage(message)
+        """Handle direct message events."""
+        message = SimpleMessage(message, self)
 
-            if _message.body.startswith("/"):
-                self.meta(_message, to=_message.sender)
-                return
+        if message.type not in self.DIRECT_MESSAGE_TYPES:
+            return
 
-            if _message.body.startswith("!"):
-                self.command(_message, to=_message.sender)
-                return
+        if message.text.startswith("@"):
+            self.command(message, to=message.sender)
 
-            try:
-                self.direct(_message)
-            except AttributeError:
-                self.log.info("Bot.direct not implemented")
+        try:
+            self.direct(message)
+        except AttributeError:
+            self.log.info(f"Bot.direct not implemented for {self.nick}")
 
     def session_start(self, message):
         """Handle session_start event."""
@@ -257,23 +302,26 @@ class Bot(ClientXMPP):
         self.plugin["xep_0045"].join_muc(message["from"], self.config.nick)
 
     def group_message(self, message):
-        """Handle groupchat_message event."""
-        if message["type"] in ("groupchat", "normal"):
-            if message["mucnick"] != self.config.nick:
-                _message = SimpleMessage(message)
+        """Handle group chat message events."""
+        message = SimpleMessage(message, self)
 
-                if _message.body.startswith("/"):
-                    self.meta(_message, room=_message.room)
-                    return
+        if message.text.startswith("@"):
+            return self.meta(message, room=message.room)
 
-                if f"{self.nick}:!" in _message.body:
-                    self.command(_message, room=_message.room)
-                    return
+        miss = message.type not in self.GROUP_MESSAGE_TYPES
+        loop = message.nick == self.nick
+        other = self.nick not in message.text
 
-                try:
-                    self.group(_message)
-                except AttributeError:
-                    self.log.info("Bot.group not implemented")
+        if miss or loop or other:
+            return
+
+        if message.content.startswith("@"):
+            self.command(message, room=message.room)
+
+        try:
+            self.group(message)
+        except AttributeError:
+            self.log.info(f"Bot.group not implemented for {self.nick}")
 
     def register_xmpp_plugins(self):
         """Register XMPP plugins that the bot supports."""
@@ -282,6 +330,17 @@ class Bot(ClientXMPP):
         self.register_plugin("xep_0199")  # XMPP Ping
         self.register_plugin("xep_0084")  # User Avatar
 
+    def init_db(self):
+        """Initialise the Redis key/value store."""
+        url = environ.get("REDIS_URL", None)
+
+        if not url:
+            self.db = None
+            self.log.info("No storage discovered")
+        else:
+            self.db = Redis.from_url(url, decode_responses=True)
+            self.log.info("Successfully connected to storage")
+
     def run(self):
         """Run the bot."""
         self.connect()
@@ -291,17 +350,17 @@ class Bot(ClientXMPP):
         except KeyboardInterrupt:
             pass
 
-    def reply(self, body, to=None, room=None):
+    def reply(self, text, to=None, room=None):
         """Send back a reply."""
         if to is None and room is None:
-            self.log("`to` or `room` arguments required for `reply`")
+            self.log.error("`to` or `room` arguments required for `reply`")
             exit(1)
 
         if to is not None and room is not None:
             self.log.error("Cannot send to both `to` and `room` for `reply`")
             exit(1)
 
-        kwargs = {"mbody": body}
+        kwargs = {"mbody": text}
         if to is not None:
             kwargs["mto"] = to
             kwargs["mtype"] = "chat"
@@ -316,28 +375,20 @@ class Bot(ClientXMPP):
         """Time since the bot came up."""
         return naturaldelta(self.start - dt.now())
 
-    def command(self, message, **kwargs):
-        """Handle "!" style commands with built-in responses."""
-        command = message.body.split("!")[-1]
+    def meta(self, message, **kwargs):
+        """Handle meta command invocations."""
+        if message.text.startswith("@bots"):
+            self.reply("🖐️", **kwargs)
 
-        if command == "uptime":
+    def command(self, message, **kwargs):
+        """Handle command invocations."""
+        if message.content.startswith("@uptime"):
             self.reply(self.uptime, **kwargs)
-        elif command == "help":
+        elif message.content.startswith("@help"):
             try:
-                self.reply(self.help, **kwargs)
+                self.reply(cleandoc(self.help), **kwargs)
             except AttributeError:
                 self.reply("No help found 🤔️", **kwargs)
-        else:
-            self.log.error(f"'{command}' command is not recognised")
-
-    def meta(self, message, **kwargs):
-        """Handle "/" style commands with built-in responses."""
-        command = message.body.split("/")[-1]
-
-        if command == "bots":
-            self.reply(f"🖐️", **kwargs)
-        else:
-            self.log.error(f"'{command}' command is not recognised")
 
 
 class EchoBot(Bot):
@@ -345,22 +396,18 @@ class EchoBot(Bot):
 
     Simply direct message the bot and see if you get back what you sent. It
     also works in group chats but in this case you need to summon the bot using
-    its nickname. Usually like so.
-
-    echobot:foo
-
+    its nickname.
     """
 
-    help = "I echo back whatever you send to me 🖖️"
+    help = "I echo messages back 🖖️"
 
     def direct(self, message):
         """Send back whatever we receive."""
-        self.reply(message.body, to=message.sender)
+        self.reply(message.text, to=message.sender)
 
     def group(self, message):
         """Send back whatever receive in group chats."""
-        if "echobot" in message.body:
-            self.reply(message.body.split(":")[-1], room=message.room)
+        self.reply(message.content, room=message.room)
 
 
 class WhisperBot(Bot):
@@ -371,11 +418,91 @@ class WhisperBot(Bot):
     to whisper your message into the group chat. The bot will then do this on
     your behalf and not reveal your identity. This is nice when you want to
     communicate with the group anonymously.
-
     """
 
-    help = "I whisper your private messages into group chats 😌️"
+    help = "I whisper private messages into group chats 😌️"
 
     def direct(self, message):
         """Receive private messages and whisper them into group chats."""
-        self.reply(f"*pssttt...* {message.body}", room=message.room)
+        self.reply(f"*pssttt...* {message.content}", room=message.room)
+
+
+class GlossBot(Bot):
+    """Building a shared glossary together.
+
+    A glossary is "an alphabetical list of terms in a particular domain of
+    knowledge with the definitions for those terms."
+
+    This bot reacts to commands which insert, list or delete items from a
+    shared glossary when summoned in a group chat. This bot makes use of
+    persistent storage so the glossary is always there even if the bot goes
+    away.
+    """
+
+    help = """
+    I help build a shared glossary
+      glossbot: @add <entry> - <definition>
+      glossbot: @rm <entry>
+      glossbot: @rand
+      glossbot: @ls
+    """
+
+    def group(self, message):
+        """Handle glossary commands."""
+        if "@add" in message.content:
+            try:
+                parsed = self.parse_add(message)
+                self.add(*parsed, room=message.room)
+            except Exception:
+                response = f"Couldn't understand '{message.content}'?"
+                self.reply(response, room=message.sender)
+        elif "@rm" in message.content:
+            try:
+                parsed = message.content.split("@rm")[-1].strip()
+                self.rm(parsed, room=message.room)
+            except Exception:
+                response = f"Couldn't understand '{message.content}'?"
+                self.reply(response, room=message.sender)
+        elif "@rand" in message.content:
+            self.rand(room=message.room)
+        elif "@ls" in message.content:
+            self.ls(room=message.room)
+        else:
+            self.log.info(f"{message.text} not recognised as glossbot command")
+
+    def parse_add(self, message):
+        """Parse the add command syntax."""
+        try:
+            replaced = message.content.replace("@add", "")
+            return [s.strip() for s in replaced.split("-")]
+        except ValueError:
+            self.log.error(f"Failed to parse {message.content}")
+
+    def add(self, entry, definition, **kwargs):
+        """Add a new entry."""
+        self.db[entry] = definition
+        self.reply("Added ✌️", **kwargs)
+
+    def rand(self, **kwargs):
+        """List a random entry."""
+        if not self.db.keys():
+            return self.reply("Glossary is empty 🙃️", **kwargs)
+
+        entry = choice(self.db.keys())
+        self.reply(f"{entry} - {self.db[entry]}", **kwargs)
+
+    def ls(self, **kwargs):
+        """List all entries."""
+        if not self.db.keys():
+            return self.reply("Glossary is empty 🙃️", **kwargs)
+
+        for entry in sorted(self.db.keys()):
+            self.reply(f"{entry} - {self.db[entry]}", **kwargs)
+
+    def rm(self, entry, **kwargs):
+        """Remove an entry."""
+        if entry not in self.db.keys():
+            return self.reply(f"{entry} doesn't exist?", **kwargs)
+
+        self.db.delete(entry)
+        self.reply("Removed ✌️", **kwargs)
